import unittest
import random
from ziplineio.app import App
from ziplineio.cache import MemoryCache, set_cache, cache
from ziplineio.dependency_injector import inject
from ziplineio.request import Request


class TestMemoryCache(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.app = App()
        set_cache(MemoryCache())

    async def test_handler_cache(self):
        @self.app.get("/cached_number")
        @cache(5)
        async def handler():
            return random.randint(0, 9999)

        req: Request = Request("GET", "/cached_number")

        first_call = await self.app._get_and_call_handler("GET", "/cached_number", req)
        second_call = await self.app._get_and_call_handler("GET", "/cached_number", req)

        # Ensure the result is cached
        self.assertEqual(first_call, second_call)

    async def test_handler_cache_with_dep_injector(self):
        class Service:
            def speak():
                return "Hello"

        @self.app.get("/cached_number")
        @inject(Service)
        @cache(5)
        async def handler(s: Service):
            return s.speak() + str(random.randint(0, 9999))

        req: Request = Request("GET", "/cached_number")

        first_call = await self.app._get_and_call_handler("GET", "/cached_number", req)
        second_call = await self.app._get_and_call_handler("GET", "/cached_number", req)

        # Ensure the result is cached
        self.assertEqual(first_call, second_call)



from contextvars import ContextVar

from ziplineio.request import Request

_request_context_var = ContextVar("request_context")


def set_request(request: Request):
    _request_context_var.set(request)


def get_request() -> Request:
    return _request_context_var.get()



from collections.abc import Callable
from typing import Any, Dict, Union
from datetime import datetime, timedelta


from ziplineio.request_context import get_request
from ziplineio.utils import call_handler


class BaseCache:
    pass

    async def get(self, key: str) -> Any:
        pass

    async def set(self, key: str, value: Any, duration: Union[int, float] = 0) -> None:
        pass

    async def clear(self) -> None:
        pass


_cache: BaseCache = None


class MemoryCache(BaseCache):
    def __init__(self):
        self._cache: Dict[str, Any] = {}
        self._expiry_times: Dict[str, datetime] = {}

    async def get(self, key: str) -> Any:
        """Get a cache entry."""
        if await self.is_expired(key):
            # remove the expired cache entry
            self._cache.pop(key, None)
            self._expiry_times.pop(key, None)
            return None
        return self._cache.get(key, None)

    async def set(self, key: str, value: Any, duration: Union[int, float] = 0) -> None:
        """Set a cache entry."""
        self._cache[key] = value
        self._expiry_times[key] = datetime.now() + timedelta(seconds=duration)

    async def is_expired(self, key: str) -> bool:
        """Check if a cache entry has expired."""
        if key not in self._expiry_times:
            return True
        return datetime.now() >= self._expiry_times[key]

    def clear(self):
        """Clears the cache."""
        self._cache.clear()
        self._expiry_times.clear()


def cache(duration: Union[int, float] = 0):
    """Cache decorator that accepts duration in seconds."""

    def decorator(func: Callable) -> Callable:
        async def wrapper(*args, **kwargs):
            req = get_request()

            url = req.path
            query_params_str = "&".join(
                [f"{k}={v}" for k, v in req.query_params.items()]
            )
            key = f"{func.__name__}:{kwargs}:{url}:{query_params_str}"

            # Check if the cache has expired or does not exist
            value = await _cache.get(key)
            if value is None:
                result = await call_handler(func, **kwargs)
                await _cache.set(key, result, duration)
                return result
            else:
                return value

        return wrapper

    return decorator


def get_cache() -> BaseCache:
    """Get the cache instance."""
    return _cache


def set_cache(cache: BaseCache) -> None:
    """Set the cache instance."""
    global _cache
    _cache = cache
    return None



from functools import wraps
from ziplineio.request import Request
from pydantic import ValidationError

from ziplineio.validation.query import QueryParam


class BodyParam(QueryParam):
    """Alias Query param."""

    pass


def validate_body(*body_params):
    """
    Decorator to validate body parameters.
    """

    def decorator(handler):
        @wraps(handler)
        async def wrapper(req: Request, *args, **kwargs):
            body = req.body.json()
            errors = {}
            validated_body = {}

            for param in body_params:
                param_name = param.param
                value = body.get(param_name)
                if value is None:
                    if param.required:
                        errors[param_name] = "Missing required body parameter"
                else:
                    try:
                        validated_body[param_name] = param.validate(value)
                    except (ValueError, ValidationError) as e:
                        errors[param_name] = str(e)

            if errors:
                return {"errors": errors}, 400

            # Attach validated body parameters to the request
            req.body = validated_body
            return await handler(req, *args, **kwargs)

        return wrapper

    return decorator if body_params else lambda req: decorator(lambda *_: None)(req)



from calendar import c
import json
from typing import Dict


class Body:
    body: bytes

    def __init__(self, body: bytes):
        self.body = body

    def get(self, key: str):
        return self.json().get(key)

    def json(self) -> Dict:
        print("body in json", self.body)
        return json.loads(self.body.decode("utf-8"))

    def __str__(self):
        return self.body.decode("utf-8")

    def bytes(self):
        return self.body

    @classmethod
    def from_json(cls, data: Dict):
        return cls(json.dumps(data).encode("utf-8"))

    @classmethod
    def from_str(cls, data: str):
        return cls(data.encode("utf-8"))

    @classmethod
    def from_bytes(cls, data: bytes):
        return cls(data)


class Request:
    def __init__(
        self,
        method: str,
        path: str,
        query_params: Dict[str, str] = {},
        path_params: Dict[str, str] = {},
        headers: Dict[str, str] = {},
        body: Body = Body(b""),
    ):
        self.method = method
        self.path = path
        self.path_params = path_params
        self.query_params = query_params
        self.headers = headers
        self.body = body

    method: str
    path: str
    query_params: Dict[str, str]
    path_params: Dict[str, str]
    headers: Dict[str, str]
    body: Body



import unittest

from ziplineio.app import App
from ziplineio.exception import BaseHttpException
from ziplineio.request import Body, Request
from ziplineio.response import (
    Response,
    StaticFileResponse,
    format_body,
    format_response,
)


class TestResponse(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.app = App()
        self.app.static("test/mocks/static", path_prefix="/static")

        @self.app.get("/")
        async def home(req):
            return {"message": "Hello, world!"}

    async def test_static_file(self):
        req = Request(method="GET", path="/static/css/test.css")
        r: StaticFileResponse = await self.app._get_and_call_handler(
            "GET", "/static/css/test.css", req
        )

        self.assertEqual(r.status, 200)
        self.assertEqual(r.get_headers()["Content-Type"], "text/css")
        self.assertTrue("background-color: #f0f0f0;" in str(r.body))


class TestFormatBody(unittest.TestCase):
    def test_format_body_dict(self):
        body = format_body({"message": "Hello, world!"})
        self.assertEqual(body, b'{"message": "Hello, world!"}')

    def test_format_body_bytes(self):
        body = format_body(b"Hello, world!")
        self.assertEqual(body, b"Hello, world!")

    def test_format_body_str(self):
        body = format_body("Hello, world!")
        self.assertEqual(body, b"Hello, world!")


class TestFormatResponse(unittest.TestCase):
    def test_format_bytes(self):
        response = b"Hello, world!"
        headers = {}
        formatted = format_response(response, headers)

        self.assertEqual(formatted["status"], 200)
        self.assertEqual(formatted["headers"], [(b"content-type", b"text/plain")])
        self.assertEqual(formatted["body"], response)

    def test_format_str(self):
        response = "Hello, world!"
        headers = {}
        formatted = format_response(response, headers)

        self.assertEqual(formatted["status"], 200)
        self.assertEqual(formatted["headers"], [(b"content-type", b"text/plain")])
        self.assertEqual(formatted["body"], b"Hello, world!")

    def test_format_dict(self):
        response = {"message": "Hello, world!"}
        headers = {}
        formatted = format_response(response, headers)

        self.assertEqual(formatted["status"], 200)
        self.assertEqual(formatted["headers"], [(b"content-type", b"application/json")])
        self.assertEqual(formatted["body"], b'{"message": "Hello, world!"}')

    def test_format_response_object(self):
        response = Response(
            200, {"Content-Type": "CUSTOM-TYPE"}, Body.from_str("Hello, world!")
        )

        formatted = format_response(response, {})
        self.assertEqual(formatted["status"], 200)
        self.assertEqual(formatted["headers"], [(b"Content-Type", b"CUSTOM-TYPE")])
        self.assertEqual(formatted["body"], b"Hello, world!")

    def test_format_exception(self):
        response = Exception("Hello, world!")
        headers = {}
        formatted = format_response(response, headers)

        self.assertEqual(formatted["status"], 500)
        self.assertEqual(formatted["headers"], [(b"content-type", b"text/plain")])
        self.assertEqual(formatted["body"], b"Internal server error: Hello, world!")

    def test_format_http_exception(self):
        response = BaseHttpException("Hello, world!", 444)
        headers = {}
        formatted = format_response(response, headers)

        self.assertEqual(formatted["status"], 444)
        self.assertEqual(formatted["headers"], [(b"content-type", b"application/json")])
        self.assertEqual(formatted["body"], b"Hello, world!")

    def test_format_bytes_with_headers(self):
        response = b"Hello, world!"
        headers = {}
        formatted = format_response(response, headers)

        self.assertEqual(formatted["status"], 200)
        self.assertEqual(formatted["headers"], [(b"content-type", b"text/plain")])
        self.assertEqual(formatted["body"], response)



class BaseHttpException(Exception):
    def __init__(self, message, status_code):
        self.message = message
        self.status_code = status_code

    def __len__(self):
        return 1


class NotFoundHttpException(BaseHttpException):
    def __init__(self, message="Not found"):
        super().__init__(message, 404)



from multiprocessing import Process
import asyncio
from jinja2 import Environment, PackageLoader, select_autoescape
import httpx
import uvicorn
import unittest


from ziplineio.app import App
from ziplineio.html.jinja import jinja
from ziplineio.request import Request
from ziplineio.router import Router
from ziplineio.service import Service

env = Environment(loader=PackageLoader("test.mocks"), autoescape=select_autoescape())

app = App()


class Service1(Service):
    name = "service1"


class Service2(Service):
    name = "service2"

    def __init__(self, service1: Service1):
        self.service1 = service1


app.inject([Service1, Service2])


@app.get("/jinja")
@jinja(env, "home.html")
def jinja_handler(service2: Service2):
    return {"content": service2.service1.name + " content"}


@app.get("/bytes")
async def bytes_handler():
    return b"Hello, world!"


@app.get("/dict")
async def dict_handler(req):
    return {"message": "Hello, world!"}


@app.get("/str")
async def str_handler():
    return "Hello, world!"


@app.post("/post")
async def post_handler(req: Request):
    body = req.body.json()

    return {"message": body}


# Will be made multithreaded
@app.get("/sync-thread")
def sync_handler():
    return {"message": "Hello, sync world!"}


user_router = Router("/user")


@user_router.get("/:id")
async def user_handler(req):
    return {"message": f"User {req.path_params['id']} received"}


@app.not_found
@jinja(env, "404.html")
def not_found():
    return {"current_route": "404"}


def run_server():
    uvicorn.run(app, port=5050)


class TestE2E(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        """Bring server up."""
        self.proc = Process(target=run_server, args=(), daemon=True)
        self.proc.start()

        # Wait for the server to be up
        await self.wait_for_server()

    async def wait_for_server(self, timeout=30):
        """Wait for the server to be ready."""

        async def check_server():
            async with httpx.AsyncClient() as client:
                while True:
                    try:
                        response = await client.get("http://localhost:5050/bytes")
                        print(response.text)
                        if response.status_code == 200:
                            return True
                    except httpx.RequestError:
                        pass
                    await asyncio.sleep(0.1)  # Short sleep between retries

        try:
            await asyncio.wait_for(check_server(), timeout=timeout)
        except asyncio.TimeoutError:
            self.fail("Server did not start within the specified timeout")

    async def asyncTearDown(self):
        self.proc.terminate()
        self.proc.join()

    async def test_handler_returns_post_body(self):
        async with httpx.AsyncClient() as client:
            response = await client.post(
                "http://localhost:5050/post", json={"message": "Hello, world!"}
            )
            print(response.json())
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), {"message": {"message": "Hello, world!"}})

    async def test_handler_returns_bytes(self):
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:5050/bytes")
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.content, b"Hello, world!")

    async def test_handler_returns_str(self):
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:5050/str")
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.text, "Hello, world!")

    async def test_handler_returns_dict(self):
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:5050/dict")
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), {"message": "Hello, world!"})

    async def test_sync_route(self):
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:5050/sync-thread")
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.json(), {"message": "Hello, sync world!"})

    async def test_jinja(self):
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:5050/jinja")
            self.assertEqual(response.status_code, 200)
            self.assertTrue("service1 content" in response.text)

    async def test_404_jinja(self):
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:5050/some-random-route")
            self.assertEqual(response.status_code, 404)
            self.assertTrue("404" in response.text)
            self.assertEqual(response.headers["Content-Type"], "text/html")



import re
import inspect
import asyncio

from typing import Dict


from ziplineio.request import Body, Request
from ziplineio.response import Response
from ziplineio.handler import Handler
from ziplineio.models import ASGIScope


"""
Only pass the kwargs that are required by the handler function.
"""


def clean_kwargs(kwargs: dict, handler: Handler) -> dict:
    params = inspect.signature(handler).parameters
    kwargs = {k: v for k, v in kwargs.items() if k in params}
    return kwargs


async def read_body(receive: callable) -> Body:
    # Read the body
    body = b""
    more_body = True

    while more_body:
        message = await receive()
        body += message.get("body", b"")
        more_body = message.get("more_body", False)

    return Body(body)


async def call_handler(
    handler: Handler,
    **kwargs,
) -> bytes | str | dict | Response | Exception:
    try:
        kwargs = clean_kwargs(kwargs, handler)
        if not inspect.iscoroutinefunction(handler):
            response = await asyncio.to_thread(handler, **kwargs)
        else:
            response = await handler(**kwargs)

    except Exception as e:
        response = e

    return response


async def parse_scope(scope: ASGIScope, receive: callable) -> Request:
    query_string = scope["query_string"].decode("utf-8")

    if query_string == "":
        query_params = {}
    else:
        query_params = dict(qp.split("=") for qp in query_string.split("&"))

    split_path = scope["path"].split(":")
    if len(split_path) > 1:
        path_params = dict(zip(split_path[1:], scope["path"].split("/")[1:]))
    else:
        path_params = {}

    body = await read_body(receive)

    headers = dict((k.decode("utf-8"), v.decode("utf-8")) for k, v in scope["headers"])
    return Request(
        method=scope["method"],
        path=scope["path"],
        query_params=query_params,
        path_params=path_params,
        headers=headers,
        body=body,
    )


def clean_url(path: str) -> str:
    return re.sub(r":\w+", "*", path)


def match_url_pattern(pattern: str, url: str) -> Dict[str, str]:
    """
    Match a URL against a pattern with wildcards and return a dictionary of captured parameters.

    Args:
        pattern (str): The pattern with wildcards (e.g., 'user/:id').
        url (str): The URL to match (e.g., 'users/123').

    Returns:
        Dict[str, str]: A dictionary of captured parameters (e.g., {'id': '123'}).
    """
    # Convert the pattern into a regular expression
    # Replace ':param' with '([^/]+)' to capture the parameter value
    regex_pattern = re.sub(r":(\w+)", r"(?P<\1>[^/]+)", pattern)
    regex_pattern = "^" + regex_pattern + "$"

    # Compile the regular expression
    pattern_re = re.compile(regex_pattern)

    # Match the URL against the pattern
    match = pattern_re.match(url)
    if not match:
        return {}

    # Return the captured parameters as a dictionary
    return match.groupdict()



import unittest


from jinja2 import Environment, PackageLoader, select_autoescape
from ziplineio.app import App
from ziplineio.request import Request

from ziplineio.html.jinja import jinja
from ziplineio.response import JinjaResponse


env = Environment(loader=PackageLoader("test.mocks"), autoescape=select_autoescape())


class TestHtml(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.app = App()

    async def test_render_jinja(self):
        @self.app.get("/")
        @jinja(env, "home.html")
        def home(req):
            return {"message": "Hello, world!"}

        req = Request(method="GET", path="/")
        response: JinjaResponse = await self.app._get_and_call_handler("GET", "/", req)

        self.assertEqual(response.status, 200)
        self.assertEqual(response.get_headers()["Content-Type"], "text/html")
        self.assertTrue("<p>Welcome to the home page!</p>" in str(response.body))



from typing import List, Callable, Tuple

from ziplineio.handler import Handler
from ziplineio.request import Request
from ziplineio.response import Response
from ziplineio.utils import call_handler


def middleware(middlewares: List[Callable]) -> Callable[[Callable], Callable]:
    def decorator(handler: Callable) -> Callable:
        async def wrapped_handler(req: Request, **kwargs):
            # Ensure 'ctx' is always present in kwargs

            kwargs.setdefault("ctx", {})
            # Run the middleware stack

            req, kwargs, res = await run_middleware_stack(
                middlewares, req=req, **kwargs
            )

            if res is not None:
                return res
            else:
                # return await handler(req, **kwargs)
                return await call_handler(handler, req=req, **kwargs)

        return wrapped_handler

    return decorator


async def run_middleware_stack(
    middlewares: list[Handler], req: Request, **kwargs
) -> Tuple[Request, dict, bytes | str | dict | Response | None]:
    for middleware in middlewares:
        # if the middleware func takes params, pass them in. Otherwise, just pass req

        if "ctx" not in kwargs:
            kwargs["ctx"] = {}

        _res = await call_handler(middleware, req=req, **kwargs)

        # regular handlers return a response, but middleware can return a tuple
        if not isinstance(_res, tuple):
            _res = (_res, kwargs)

        if len(_res) != 2:
            req = _res
            middleware_ctx = kwargs["ctx"]
        else:
            req, middleware_ctx = _res
            kwargs["ctx"].update(middleware_ctx)

        if not isinstance(req, Request):
            response = req
            return req, kwargs, response

    return req, kwargs, None
